"""
Module to provide the TelemetryData classes

Dataclasses defined in Module

- TelemetryData: Data class to hold the actual (error)counters in a
                 TelemetryModel object

- TelemetryModel: Data class to define the Telemetry object category, type,
                  source etc. Holds the reference to the actual counters defined
                  in TelemetryData dataclass.
"""

from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from typing import DefaultDict, Dict, Optional

from errors import ErrorCode

from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings import settings as st


@dataclass
class TelemetryData:
    """
    Class to define the telemetry (error)counters.

    attributes:
    - base_counter (int): counter for the TelemetryData object.
    - fail_counter (int): fail counter for the TelemetryData object.
    - counters (dict): dict of sub (custom) counters [str: int]
    - errors (dict): dict of error counters [str: int]

    public methods:
    - increase_base_count
    - increase_fail_count
    - increase_custom_count
    - increase_error_count

    counters can be added in which case base, fail, custom and error counters
    will be summed up seperately. Add method will return self with added
    TelemetryData object.
    """

    base_counter: int = 0
    fail_counter: int = 0
    counters: DefaultDict[str, int] = field(default_factory=lambda: defaultdict(int))
    errors: DefaultDict[str, int] = field(default_factory=lambda: defaultdict(int))

    def increase_base_count(self, increment: int) -> None:
        """Increase the base counter with a given increment."""
        self.base_counter += increment

    def increase_fail_count(self, increment: int) -> None:
        """Increase the fail counter with a given increment."""
        self.fail_counter += increment

    def increase_error_count(self, increment: int, error_code: ErrorCode) -> None:
        """Increase an error counter with a given increment."""

        error_code_key = error_code.code
        self._increase_error_count(increment=increment, error_code_key=error_code_key)

    def _increase_error_count(self, increment: int, error_code_key: str) -> None:
        self.errors[error_code_key] += increment

    def increase_custom_count(self, increment: int, counter: str) -> None:
        """Increase a custom counter with a given increment."""
        self.counters[counter] += increment

    def __add__(self, telemetry_data: "TelemetryData") -> "TelemetryData":
        """
        Add telemetry_data object to self by adding up all counters seperately.
        """
        self.increase_base_count(telemetry_data.base_counter)
        self.increase_fail_count(telemetry_data.fail_counter)
        for error_code_key, increment in telemetry_data.errors.items():
            self._increase_error_count(
                increment=increment, error_code_key=error_code_key
            )
        for counter, increment in telemetry_data.counters.items():
            self.increase_custom_count(increment=increment, counter=counter)
        return self


@dataclass
class TelemetryModel:
    telemetry_type: str
    category: str
    sub_category: str
    source_name: str
    process_type: str
    start_date_time: datetime = datetime.now()
    run_time_in_seconds: Optional[float] = None
    io_time_in_seconds: float = 0
    traffic_light: str = st.DEFAULT_TRAFIC_LIGHT_COLOR
    telemetry: Dict[str, TelemetryData] = field(default_factory=dict)

    def validate(self) -> None:
        self._check_telemetry_type()

    def copy(self) -> "TelemetryModel":
        """
        Method to return a copy of the telemetry model. In a telemetry copy
        only the attributes telemetry_type, categroy, sub_category, source_name
        and process_type are copied.
        """
        return TelemetryModel(
            telemetry_type=self.telemetry_type,
            category=self.category,
            sub_category=self.sub_category,
            source_name=self.source_name,
            process_type=self.process_type,
        )

    def _check_telemetry_type(self) -> None:
        """Check validaty of provided telemetry type.

        Raises exception if not valid. Returns non if telemetry type is valid.

        Args:
            telemetry_type (str): [description]

        Raises:
            exceptions.InvalidTelemetryType: When telemetry type is not valid.
        """
        if self.telemetry_type not in st.TELEMETRY_TYPES:
            raise exceptions.InvalidTelemetryType(st.TELEMETRY_TYPES)

    def set_orange_traffic_light(self) -> None:
        """Sets traffic light attribute to orange."""
        self.traffic_light = st.TRAFIC_LIGHT_COLOR_ORANGE

    def set_red_traffic_light(self) -> None:
        """Sets traffic light attribute to red."""
        self.traffic_light = st.TRAFIC_LIGHT_COLOR_RED

    def __add__(self, telemetry_model_to_add: "TelemetryModel") -> "TelemetryModel":
        """
        Method to add to telementry model instances.
        Adding a 2 telemetry model instances implies adding all telemetry data
        objects and adding iotime, run time and traffic light attributes to a
        specific counter.
        """
        self.__add_base_count()
        self.__add_sub_process(telemetry_model_to_add=telemetry_model_to_add)
        self.__add_traffic_light(telemetry_model_to_add=telemetry_model_to_add)
        self.__add_io_time(telemetry_model_to_add=telemetry_model_to_add)
        self.__add_run_time(telemetry_model_to_add=telemetry_model_to_add)
        return self

    def __add_base_count(self) -> None:
        """
        Sub method for the __add__ method to increase base_counter of the
        aggregation sub_process when adding TelemetryModel instances.
        """
        self.get_sub_process_data(sub_process=st.AGGREGATION_KEY).increase_base_count(
            increment=1
        )

    def __add_traffic_light(self, telemetry_model_to_add: "TelemetryModel") -> None:
        """
        Sub method for the __add__ method to add the traffic_light value of the
        TelemetryModel instance to be added to the aggregation sub_process.
        """
        self.get_sub_process_data(sub_process=st.AGGREGATION_KEY).increase_custom_count(
            increment=1, counter=telemetry_model_to_add.traffic_light
        )

    def __add_sub_process(self, telemetry_model_to_add: "TelemetryModel") -> None:
        """
        Sub method for the __add__ method to add the sub_processes of the two
        TelemetryModel instances.
        """
        for sub_process in telemetry_model_to_add.telemetry:
            sub_pr = self.get_sub_process_data(sub_process)
            sub_pr += telemetry_model_to_add.get_sub_process_data(
                sub_process=sub_process
            )

    def __add_io_time(self, telemetry_model_to_add: "TelemetryModel") -> None:
        """
        Sub method for the __add__ method to add the rounded io_time of the
        TelemetryModel instance to be added to the aggregation sub_process.
        """
        self.get_sub_process_data(sub_process=st.AGGREGATION_KEY).increase_custom_count(
            increment=round(telemetry_model_to_add.io_time_in_seconds),
            counter=st.IO_TIME_KEY,
        )

    def __add_run_time(self, telemetry_model_to_add: "TelemetryModel") -> None:
        """
        Sub method for the __add__ method to add the rounded run_time of the
        TelemetryModel instance to be added to the aggregation sub_process.
        """
        run_time_in_seconds = telemetry_model_to_add.run_time_in_seconds or 0
        self.get_sub_process_data(sub_process=st.AGGREGATION_KEY).increase_custom_count(
            increment=round(run_time_in_seconds), counter=st.RUN_TIME
        )

    def get_sub_process_data(self, sub_process: str) -> TelemetryData:
        if not self.telemetry.get(sub_process):
            self.telemetry[sub_process] = TelemetryData()
        return self.telemetry[sub_process]



extensions = [
    # "sphinx.ext.autodoc",
    # "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel"
]
source_suffix = ".rst"
master_doc = "index"
project = "Pipeline Telemetry"
year = "2024"
author = "Maarten de Ruyter"
copyright = "{0}, {1}".format(year, author)
version = release = "1.0.1"

extensions = []

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']

html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
}
html_short_title = "%s-%s" % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



"""
Helper methods provide date range iterarors.

These date ranges are to be used in TelemetryStorage queries used
by the TelemetryAggregator classes.

The date range includes start_date and end_date to be used in the query.

generic methods:

- date_range_generator
        generic generator, not likely to be used outside this module

Single day date range generators.
- get_daily_date_ranges: free choice of start and end date
- get_daily_date_range_till_yesterday: from start date till yesterday
- get_daily_date_range_yesterday: date range for yesterday


"""

from datetime import datetime
from typing import NamedTuple, Protocol

from pipeline_telemetry.data_classes import TelemetryModel


class TelemetrySelector(NamedTuple):
    """Named tuple to define mandatory attributed for a Telemetry selection.

    This object should be used as input to the Telemetry Aggregator

    >>> telemetry_selector = TelemetrySelector(
        category, sub_category, source_name, process_type)
    >>> aggregator = AggregatorClass(
            telemetry_selector=telemetry_selector,
            telemetry_storage=TelemetryStorageClass()
        )

    The `telemtry_selector` defines what telemetry objects are in scope.
    The `aggregator` object defines what kind of aggregation is done (by
    selecting the AggregatorClass) and provides you with metods to run the
    aggergation for a specific period.
    """

    category: str
    sub_category: str
    source_name: str
    process_type: str


class TelemetryListArgs(NamedTuple):
    telemetry_type: str
    category: str
    sub_category: str
    source_name: str
    process_type: str
    from_date_time: datetime
    to_date_time: datetime


class TelemetryList(Protocol):
    def __next__(self) -> TelemetryModel: ...

    def __iter__(self) -> "TelemetryList": ...


class TelemetryAggregator:
    __telemetry: TelemetryModel

    def __init__(self, telemetry: TelemetryModel) -> None:
        self.__telemetry = telemetry

    def aggregate(self, telemetry_list: TelemetryList) -> TelemetryModel:
        for telemetry in telemetry_list:
            self.__telemetry += telemetry
        return self.__telemetry



"""Module to define test data for the integration tests"""

from dataclasses import dataclass

from pipeline_telemetry import Telemetry
from pipeline_telemetry.data_classes import TelemetryData
from pipeline_telemetry.settings import settings as st


@dataclass(frozen=True)
class IntegrationTestData:
    """Immutable dataclass to define integration test data."""

    data: dict
    validation_rules: dict
    expected_result: TelemetryData
    result_field: str

    def validate_telemetry_data(self, telemetry: Telemetry) -> bool:
        test_result = telemetry.get(self.result_field)
        expected_result = self.expected_result

        return test_result == expected_result

        # return \
        #     test_result.base_counter == expected_result.base_counter and \
        #     test_result.fail_counter == expected_result.fail_counter and \
        #     test_result.errors == expected_result.errors and \
        #     test_result.counters == expected_result.counters


BASIC_RETRIEVE_DATA_TEST = IntegrationTestData(
    data={
        "items": [
            {"id": 1, "field1": "test"},
            {"id": 2, "field1": "test"},
            {"id": 3, "field1": "test"},
            {"id": 4, "field1": "test"},
        ]
    },
    validation_rules={
        "RETRIEVE_RAW_DATA": {
            "has_key": {"field_name": "items"},
            "validate_entries": {"field_name": "items", "expected_count": 4},
        }
    },
    result_field="RETRIEVE_RAW_DATA",
    expected_result=TelemetryData(**{"base_counter": 1, "fail_counter": 0}),
)

MISSING_KEY_RETRIEVE_DATA_TEST = IntegrationTestData(
    data={"no items key": []},
    validation_rules={"RETRIEVE_RAW_DATA": {"has_key": {"field_name": "items"}}},
    result_field="RETRIEVE_RAW_DATA",
    expected_result=TelemetryData(
        **{
            "base_counter": 1,
            "fail_counter": 0,
            st.ERRORS_KEY: {"HAS_KEY_ERR_0001@KEY_<items>": 1},
        }
    ),
)


MUST_HAVE_KEY_IN_ITEMS_TEST = IntegrationTestData(
    data={"items": [{"key": 1}, {"no_key": 2}]},
    validation_rules={
        "RETRIEVE_RAW_DATA": {
            "entries_have_key": {"field_name": "items", "must_have_key": "key"}
        }
    },
    result_field="RETRIEVE_RAW_DATA",
    expected_result=TelemetryData(
        **{
            "base_counter": 1,
            "fail_counter": 0,
            st.ERRORS_KEY: {"ENTRIES_HAVE_KEY_ERR_003@KEY_<items__key>": 1},
        }
    ),
)

NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST = IntegrationTestData(
    data={"items": [{"no_key": 1}, {"no_key": 2}]},
    validation_rules={
        "RETRIEVE_RAW_DATA": {
            "entries_have_key": {"field_name": "items", "must_have_key": "key"}
        }
    },
    result_field="RETRIEVE_RAW_DATA",
    expected_result=TelemetryData(
        **{
            "base_counter": 1,
            "fail_counter": 0,
            st.ERRORS_KEY: {"ENTRIES_HAVE_KEY_ERR_003@KEY_<items__key>": 2},
        }
    ),
)


ITEMS_FIELD_HAS_DICT_TEST = IntegrationTestData(
    data={"items": {"key": 1, "no_key": 2}},
    validation_rules={
        "RETRIEVE_RAW_DATA": {
            "entries_have_key": {"field_name": "items", "must_have_key": "key"}
        }
    },
    result_field="RETRIEVE_RAW_DATA",
    expected_result=TelemetryData(
        **{
            "base_counter": 1,
            "fail_counter": 0,
            st.ERRORS_KEY: {"ENTRIES_HAVE_KEY_ERR_002@KEY_<items>": 1},
        }
    ),
)



""" """

from .dict_validator import DictValidator
from .entries_have_key import EntriesHaveKey
from .has_key import HasKey
from .validate_entries import ValidateEntries

__all__ = ["DictValidator", "EntriesHaveKey", "HasKey", "ValidateEntries"]



"""
Module to test telemetry counter login for pipeline telemetry module.
"""

import pytest
import test_data as TD

from pipeline_telemetry import ProcessTypes
from pipeline_telemetry.helper import is_telemetry_counter
from pipeline_telemetry.main import Telemetry
from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings import settings as st
from pipeline_telemetry.settings.data_class import TelemetryCounter


@pytest.fixture
def telemetry_inst():
    telemetry_inst = Telemetry(**TD.DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    return telemetry_inst


def test_telemetry_counter_exists():
    """check that TelemetryCounter class exists"""
    assert TelemetryCounter


def test_add_to_method():
    """
    Test that add_method on a telemetry counter adds the telemetry counter
    to given object with a _telemetry attribute
    """
    sub_process = TD.TEST_TELEMETRY_COUNTER_INC_2.sub_process
    telemetry = Telemetry(**TD.DEFAULT_TELEMETRY_PARAMS)

    class ClassToTestAddTo:
        _telemetry = telemetry

    obj_with_telemetry = ClassToTestAddTo()

    # ensure sub process is initialized in object's telemetry
    telemetry._initialize_sub_process(sub_process)

    # run add_to method on a Telemetry counter
    TD.TEST_TELEMETRY_COUNTER_INC_2.add_to(obj_with_telemetry)
    assert sub_process in obj_with_telemetry._telemetry.telemetry_data
    telemetry_data = obj_with_telemetry._telemetry.get(sub_process)
    assert getattr(telemetry_data, st.COUNTERS_KEY)["test_counter"] == 2


def test_add_telemetry_counter_to_telemetry(telemetry_inst):
    """
    Check that adding a telemetry counter to a Telemetry instance
    creates a counter
    """
    telemetry_inst.add_telemetry_counter(TD.TEST_TELEMETRY_COUNTER)
    telemetry_data = telemetry_inst.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, st.COUNTERS_KEY)["test_counter"] == 1


def test_add_error_telemetry_counter_to_telemetry(telemetry_inst):
    """
    Check that adding a telemetry counter to a Telemetry instance
    creates a counter
    """
    telemetry_inst.add_telemetry_counter(TD.TEST_ERROR_TELEMETRY_COUNTER)
    telemetry_data = telemetry_inst.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, st.ERRORS_KEY)["HAS_KEY_ERR_0001"] == 1


def test_add_telemetry_counter_when_sub_process_is_not_initialized():
    """
    Check that adding a telemetry counter to a Telemetry instance possible when
    subproces is not initialized
    """
    telemetry_inst = Telemetry(**TD.DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.add_telemetry_counter(TD.TEST_TELEMETRY_COUNTER)
    telemetry_data = telemetry_inst.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, st.COUNTERS_KEY)["test_counter"] == 1


def test_add_telemetry_counter_with_icrement_2_to_telemetry(telemetry_inst):
    """
    Check that adding a telemetry counter with increment = 2 to a Telemetry
    raises the counter with value of 2
    """
    telemetry_inst.add_telemetry_counter(TD.TEST_TELEMETRY_COUNTER)
    telemetry_data = telemetry_inst.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, st.COUNTERS_KEY)["test_counter"] == 1
    telemetry_inst.add_telemetry_counter(TD.TEST_TELEMETRY_COUNTER_INC_2)
    assert getattr(telemetry_data, st.COUNTERS_KEY)["test_counter"] == 3


def test_add_telemetry_counter_with_increment_arg(telemetry_inst):
    """
    Check invoking add_telemetry_counter method using increment argument adds
    the value of increment to the counter
    """
    telemetry_inst.add_telemetry_counter(TD.TEST_TELEMETRY_COUNTER, increment=10)
    telemetry_data = telemetry_inst.get("RETRIEVE_RAW_DATA")
    assert getattr(telemetry_data, st.COUNTERS_KEY)["test_counter"] == 10


def test_add_telemetry_counter_raises_exception(telemetry_inst):
    """
    Check invoking add_telemetry_counter method using increment argument adds
    the value of increment to the counter
    """
    with pytest.raises(exceptions.InvalidSubProcessForProcessType):
        telemetry_inst.add_telemetry_counter(TD.TEST_INV_TELEMETRY_COUNTER)


def test_is_telemetry_counter_returns_false():
    """
    Test is_telemetry_counter method returns False if counter is not an instance
    off TelemetryCounter or subclass of TelemetryCounter.
    """
    assert not is_telemetry_counter("1")


def test_is_telemetry_counter_returns_true():
    """
    Test is_telemetry_counter method returns True if counter is an instance of
    TelemetryCounter.
    """
    assert is_telemetry_counter(TD.TEST_TELEMETRY_COUNTER)


def test_is_error_returns_true_on_error_code_sub_class():
    """
    Test is_error method returns True if error is an instance of ErrorCode.
    """

    class TelemetryCounterSubClass(TelemetryCounter):
        pass

    counter = TelemetryCounterSubClass(
        process_type=TD.TEST_TELEMETRY_COUNTER.process_type,
        sub_process=TD.TEST_TELEMETRY_COUNTER.sub_process,
        counter_name=TD.TEST_TELEMETRY_COUNTER.counter_name,
    )

    assert is_telemetry_counter(counter)


def test_telemetry_counter_validate_sub_process():
    """
    Test validate_sub_process method returns None if sub process is in scope of
    process_type.
    """
    assert TD.TEST_TELEMETRY_COUNTER.validate_sub_process() is None


def test_validate_sub_process_raises_exception():
    """
    Test validate_sub_process method raises exception if sub process is not
    in scope of process_type.
    """
    counter = TelemetryCounter(
        process_type=TD.TEST_TELEMETRY_COUNTER.process_type,
        sub_process="sub process not in scope of process_type",
        counter_name=TD.TEST_TELEMETRY_COUNTER.counter_name,
    )

    with pytest.raises(exceptions.InvalidSubProcessForProcessType):
        counter.validate_sub_process()


def test_validate_sub_process_with_multiple_process_types():
    """
    Test validate_sub_process method returns None if sub process is in scope of
    one of the given process_types.
    """
    counter = TelemetryCounter(
        process_types=[ProcessTypes.CREATE_DATA_FROM_URL, ProcessTypes.UPLOAD_DATA],
        sub_process="DATA_UPLOAD",
        counter_name=TD.TEST_TELEMETRY_COUNTER.counter_name,
    )
    assert counter.validate_sub_process() is None


def test_all_process_types_with_multiple_entries():
    """
    Test validate_sub_process method returns None if sub process is in scope of
    one of the given process_types.
    """
    counter = TelemetryCounter(
        process_types=[ProcessTypes.CREATE_DATA_FROM_URL, ProcessTypes.UPLOAD_DATA],
        sub_process="DATA_UPLOAD",
        counter_name=TD.TEST_TELEMETRY_COUNTER.counter_name,
    )
    assert None not in counter.all_process_types
    assert ProcessTypes.CREATE_DATA_FROM_URL in counter.all_process_types
    assert ProcessTypes.UPLOAD_DATA in counter.all_process_types


def test_all_process_types_with_single_entry():
    """
    Test validate_sub_process method returns None if sub process is in scope of
    one of the given process_types.
    """
    counter = TelemetryCounter(
        process_type=ProcessTypes.CREATE_DATA_FROM_URL,
        sub_process="DATA_UPLOAD",
        counter_name=TD.TEST_TELEMETRY_COUNTER.counter_name,
    )
    assert None not in counter.all_process_types
    assert ProcessTypes.CREATE_DATA_FROM_URL in counter.all_process_types


def test_set_increment_method_on_telemetry_counter():
    """
    Test set_increment sets a new value for the increment.
    """
    counter = TelemetryCounter(
        process_type=ProcessTypes.CREATE_DATA_FROM_URL,
        sub_process="DATA_UPLOAD",
        counter_name=TD.TEST_TELEMETRY_COUNTER.counter_name,
    )
    assert counter.increment == 1
    new_counter = counter.set_increment(increment=10)
    assert new_counter.increment == 10


def test_telementry_counter_hash():
    """Test that a TelemetryCounter object can be hashed."""
    assert TD.TEST_TELEMETRY_COUNTER.__hash__()


def test_tc_hash_does_include_the_increment():
    """Test that increment value changes the hash."""
    assert (
        not TD.TEST_TELEMETRY_COUNTER.__hash__()
        == TD.TEST_TELEMETRY_COUNTER_INC_2.__hash__()
    )


def test_tc_hash_works_with_errors():
    """
    Test that a telementry counters with different errors return a different
    hash.
    """
    assert (
        not TD.TEST_ERROR_TELEMETRY_COUNTER.__hash__()
        == TD.TEST_ERROR_TELEMETRY_COUNTER_2.__hash__()
    )


def test_tc_hash_works_with_counters():
    """
    Test that a telementry counters with different counters return a different
    hash.
    """
    assert (
        not TD.TEST_TELEMETRY_COUNTER.__hash__()
        == TD.TEST_TELEMETRY_COUNTER_2.__hash__()
    )


def test_tc_hash_does_include_process_types_field():
    """
    Test that a telementry counters with a different list of process types
    return a different hash.
    """
    assert (
        not TD.TEST_TC_MULT_PROCESS_TYPES.__hash__()
        == TD.TEST_TC_MULT_PROCESS_TYPES_2.__hash__()
    )


def test_tc_hash_does_include_sub_process_field():
    """
    Test that a telementry counters with a different sub_process
    return a different hash.
    """
    assert (
        not TD.TEST_TELEMETRY_COUNTER.__hash__()
        == TD.TEST_TELEMETRY_COUNTER_3.__hash__()
    )



"""Module to define EntriesHaveKey class validator"""

from dataclasses import dataclass

import jmespath
from errors import ErrorCode, ListErrors

from .abstract_validator_instruction import (
    AbstractValidatorInstruction,
    BaseValidatorInstructionRuleData,
)


@dataclass(frozen=True)
class EntriesHaveKeyRuleData(BaseValidatorInstructionRuleData):
    must_have_key: str


class EntriesHaveKey(AbstractValidatorInstruction):
    """
    class to define EntriesHaveKey instruction. Validator checks if all
    entries in a list are of type dict and contain a specific key/field.

    Validator can be invoked with command `entries_have_key`
    rule_data needs to contain:
        - field_name (str): field that holds list of entries to be
                            validated
        - must_have_key (str): field that hold the key that is to be present
                               in all entries

    public method
        - validate
    """

    RULE_DATA_CLASS = EntriesHaveKeyRuleData
    INSTRUCTION = "entries_have_key"
    must_have_key = "must_have_key"

    @classmethod
    def _validate(
        cls, dict_to_validate: dict, rule_data: BaseValidatorInstructionRuleData
    ) -> list[ErrorCode]:
        """
        method to do the actual validation

        returns:
            - list of errorcodes
        """
        # if entries does not exist stop validation
        if error := cls._validate_field_exists(dict_to_validate, rule_data):
            return error

        # if entries field is not a list stop validation
        if error := cls._validate_type_entries_field(dict_to_validate, rule_data):
            return error

        # return the errors per entry
        return cls._validate_entries_have_key(dict_to_validate, rule_data)

    @classmethod
    def _validate_entries_have_key(
        cls, dict_to_validate: dict, rule_data: BaseValidatorInstructionRuleData
    ) -> list[ErrorCode]:
        """
        method to check if the values in the entry have a specific key

        returns:
            - list{ErrorCode]: Empty list in case all entries have the key
        """
        errors = []
        fieldname = cls._get_field_name(rule_data)
        must_have_key = cls._get_must_have_key(rule_data)

        # prepare error messages
        entry_is_not_a_dict_error = cls._validation_error(
            ListErrors.ENTRY_IS_NOT_A_DICT, fieldname
        )
        missing_key_in_entry_error = cls._validation_error(
            ListErrors.KEY_NOT_FOUND_IN_ENTRY, fieldname + "__" + must_have_key
        )

        # for each entry do a has_key validation with the `must_have_key` field
        for entry in jmespath.search(fieldname, dict_to_validate):
            if not isinstance(entry, dict):
                errors.extend(entry_is_not_a_dict_error)
                continue

            if not jmespath.search(must_have_key, entry):
                errors.extend(missing_key_in_entry_error)

        return errors

    @classmethod
    def _validate_field_exists(
        cls, dict_to_validate: dict, rule_data: BaseValidatorInstructionRuleData
    ) -> list[ErrorCode]:
        """
        method to check if entries field exists in dict_to_validate

        returns:
            - list{ErrorCode]: Empty list in case the field does exist
        """
        fieldname = cls._get_field_name(rule_data)

        if not jmespath.search(fieldname, dict_to_validate):
            return cls._validation_error(ListErrors.ENTRIES_FIELD_NOT_FOUND, fieldname)

        return []

    @classmethod
    def _validate_type_entries_field(
        cls, dict_to_validate: dict, rule_data: BaseValidatorInstructionRuleData
    ) -> list[ErrorCode]:
        """
        method to check if entries field in dict_to_validate is a list

        returns:
            - list[ErrorCode]: Empty list in case the field holds a list
        """
        fieldname = cls._get_field_name(rule_data)
        entries = jmespath.search(fieldname, dict_to_validate)
        if not isinstance(entries, (list)):
            return cls._validation_error(
                ListErrors.ENTRIES_FIELD_OF_WRONG_TYPE, fieldname
            )

        return []

    @classmethod
    def _get_must_have_key(cls, rule_data: BaseValidatorInstructionRuleData) -> str:
        """
        retrieves the must_have_key from the rule content
        """
        return getattr(rule_data, cls.must_have_key)

    @staticmethod
    def _validation_error(error_code: ErrorCode, fieldname: str) -> list[ErrorCode]:
        """returns error code object in list with fieldname as error data"""
        key_specific_error = error_code.code + "@KEY_<" + fieldname + ">"
        return [ErrorCode(code=key_specific_error, description=error_code.description)]



"""
module to test the class methods fro Telemetry class of telemetry module

"""

import pytest

from pipeline_telemetry.main import Telemetry
from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings.data_class import ProcessType


def test_telemetry_class_method_add_process_type():
    """
    check that Telemetry class method add_process_type adds a process type
    """
    process_type = ProcessType(
        process_type="custom_process_type", subtypes=["custom_sub_process_type"]
    )
    Telemetry.add_process_type("TEST_TYPE", process_type)
    assert Telemetry("test", "test", "test", process_type)


def test_sub_process_types_also_added():
    """
    check that Telemetry class method add_process_type adds a process type
    including all sub_processes.
    """
    process_type = ProcessType(
        process_type="custom_process_type", subtypes=["custom_sub_process_type"]
    )
    Telemetry.add_process_type("TEST_TYPE", process_type)
    telemetry = Telemetry("category", "sub_category", "name", process_type)
    assert telemetry._process_type.sub_processes == ["custom_sub_process_type"]


def test_add_process_type_raises_exception():
    """
    test class method add_process_type raises an excpetion when an object
    other the a ProcessType is offered
    """
    with pytest.raises(exceptions.ProcessTypeMustBeOfClassProcessType):
        Telemetry.add_process_type("TYPE_NAME", "not a ProcessType")



"""
This module provides test data for the telemetry tests
"""

from errors import ErrorCode, ListErrors, ReturnValueWithStatus

from pipeline_telemetry import ProcessType, ProcessTypes, TelemetryCounter
from pipeline_telemetry.settings.settings import DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES
from pipeline_telemetry.validators.abstract_validator_instruction import (
    AbstractValidatorInstruction,
)


class InstructionTestClass(AbstractValidatorInstruction):
    """Instruction class for test purposes."""

    INSTRUCTION = "test_instruction"


# default params for creating a test Telemetry object
DEFAULT_TELEMETRY_PARAMS = {
    "category": "WEATHER",
    "sub_category": "DAILY_PREDICTIONS",
    "source_name": "load_weather_data",
    "process_type": ProcessTypes.CREATE_DATA_FROM_URL,
}

DEFAULT_TELEMETRY_MODEL_PARAMS = {
    "telemetry_type": "SINGLE TELEMETRY",
    "category": "WEATHER",
    "sub_category": "DAILY_PREDICTIONS",
    "source_name": "load_weather_data",
    "process_type": ProcessTypes.CREATE_DATA_FROM_URL.name,
}

# custom process type for testing extending the default process types
CUSTOM_PROCESS_TYPE = {"custom_process_type": ["CUSTOM_SUB_TYPE"]}

TEST_ERROR_CODE = ErrorCode(code="TEST_CODE_0001", description="test error")
TEST_ERROR_CODE_2 = ErrorCode(code="TEST_CODE_0002", description="test error")

TEST_TELEMETRY_RULES = {"RETRIEVE_RAW_DATA": {"has_key": {"field_name": "items"}}}

TEST_PROCESS_TYPE = ProcessType(
    process_type="test_process_type", subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES
)


TEST_PROCESS_TYPE_2 = ProcessType(
    process_type="test_process_type", subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES
)

TEST_PROCESS_TYPE_3 = ProcessType(
    process_type="test_process_type_3", subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES
)

TEST_TELEMETRY_COUNTER = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="RETRIEVE_RAW_DATA",
    counter_name="test_counter",
)

TEST_TELEMETRY_COUNTER_2 = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="RETRIEVE_RAW_DATA",
    counter_name="test_counter_2",
)

TEST_TELEMETRY_COUNTER_3 = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="DATA_CONVERSION",
    counter_name="test_counter",
)


TEST_TELEMETRY_COUNTER_INC_2 = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="RETRIEVE_RAW_DATA",
    counter_name="test_counter",
    increment=2,
)

TEST_INV_TELEMETRY_COUNTER = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="invalid_sub_process",
    counter_name="test_counter",
    increment=2,
)

TEST_ERROR_TELEMETRY_COUNTER = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="RETRIEVE_RAW_DATA",
    increment=1,
    error=ListErrors.KEY_NOT_FOUND,
)

TEST_ERROR_TELEMETRY_COUNTER_2 = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="RETRIEVE_RAW_DATA",
    increment=1,
    error=ListErrors.FIELD_NOT_FOUND,
)


TEST_ERROR_TC_PROCESS_TYPE_2 = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="RETRIEVE_RAW_DATA",
    increment=1,
    error=ListErrors.KEY_NOT_FOUND,
)

TEST_TC_MULT_PROCESS_TYPES = TelemetryCounter(
    process_types=[TEST_PROCESS_TYPE, TEST_PROCESS_TYPE_2],
    sub_process="RETRIEVE_RAW_DATA",
    counter_name="test_counter",
)

TEST_TC_MULT_PROCESS_TYPES_2 = TelemetryCounter(
    process_types=[TEST_PROCESS_TYPE],
    sub_process="RETRIEVE_RAW_DATA",
    counter_name="test_counter",
)


TEST_RESULT = [1, TEST_TELEMETRY_COUNTER, 2, TEST_ERROR_TELEMETRY_COUNTER]

TEST_RETURN_VALUE = ReturnValueWithStatus(result=TEST_RESULT)



"""Module to define abstract validator class"""

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import List, Type

from errors import ErrorCode


@dataclass(frozen=True)
class BaseValidatorInstructionRuleData:
    field_name: str

    def __post_init__(self):
        if not isinstance(self.field_name, str):
            raise TypeError("Field 'field_name' must be of type 'str'.")


class AbstractValidatorInstruction(metaclass=ABCMeta):
    """Abstract Validator Instruction class"""

    RULE_DATA_CLASS: Type[BaseValidatorInstructionRuleData] = (
        BaseValidatorInstructionRuleData
    )
    INSTRUCTION = "instruction_name"
    FIELDNAME = "field_name"

    def __new__(cls):
        """make this a singleton class"""
        return cls

    @classmethod
    def validate(cls, dict_to_validate: dict, rule_dict: dict) -> list[ErrorCode]:
        """
        Public method to run the validation
        prior to validation the rule will be validated.

        Args:
            dict_to_validate (dict): Input dict that needs validation
            rule_content (dict): actual rule

        Returns:
            list[ErrorCode]: List with errors that were found

        Raise:
            FieldNameMandatory: when field name not defined in rule
        """
        rule_data = cls.RULE_DATA_CLASS(**rule_dict)
        return cls._validate(dict_to_validate, rule_data)

    @classmethod
    @abstractmethod
    def _validate(
        cls, dict_to_validate: dict, rule_data: BaseValidatorInstructionRuleData
    ) -> List[ErrorCode]:
        """
        method to do the actual validation

        returns:
            - list of errorcodes
        """

    @classmethod
    def _get_field_name(cls, rule_data: BaseValidatorInstructionRuleData) -> str:
        """
        Retrieves the field name in scope from the rule content.
        """
        return getattr(rule_data, cls.FIELDNAME)

    @staticmethod
    @abstractmethod
    def _validation_error(error_code: ErrorCode, fieldname: str) -> list[ErrorCode]:
        """
        Customize the errors and their content for the specific validation rule.
        """



""" """

from datetime import date, datetime, timedelta
from typing import Iterator, NamedTuple


class DateRange(NamedTuple):
    from_date: date
    to_date: date


class DateTimeRange(NamedTuple):
    from_date: datetime
    to_date: datetime


def date_range_to_date_time_range(date_range: DateRange) -> DateTimeRange:
    start_date, end_date = date_range
    start_date_time = datetime(*start_date.timetuple()[:6])
    end_date_time = datetime(*end_date.timetuple()[:6])
    return DateTimeRange(start_date_time, end_date_time)


def date_range_generator(
    start_date_time: datetime, end_date_time: datetime, time_delta: timedelta
) -> Iterator[DateTimeRange]:
    """Iterator for a list of DateTimeRange object

    Args:
        start_date_time (datetime): start date
        end_date_time (datetime): end date
        time_delta (timedelta): lengt of date time range

    The resulting date
    Yields:
        Iterator[DateTimeRange]: _description_
    """
    start_date = start_date_time.date()
    end_date = start_date + time_delta
    while end_date <= end_date_time.date():
        yield date_range_to_date_time_range(DateRange(start_date, end_date))
        start_date, end_date = end_date, end_date + time_delta


def get_daily_date_ranges(start_date: date, end_date: date) -> Iterator[DateTimeRange]:
    """
    Iterator to return single day date ranges from start_date to end_date.
    """
    start_date_time = datetime(*start_date.timetuple()[:6])
    end_date_time = datetime(*end_date.timetuple()[:6])
    return date_range_generator(
        start_date_time=start_date_time,
        end_date_time=end_date_time,
        time_delta=timedelta(days=1),
    )


def get_daily_date_range_till_yesterday(start_date: date) -> Iterator[DateTimeRange]:
    """
    Iterator to return single day date ranges from start_date till yesterday.
    """
    end_date = date.today()
    return get_daily_date_ranges(start_date=start_date, end_date=end_date)


def get_daily_date_range_yesterday() -> Iterator[DateTimeRange]:
    """
    Iterator to return single day date ranges for yesterday.
    """
    end_date = date.today()
    start_date = date.today() - timedelta(days=1)
    return get_daily_date_ranges(start_date=start_date, end_date=end_date)


def get_daily_date_range_for_single_date(date: date) -> DateTimeRange:
    """
    Return single daily date range for a given date.
    """
    return next(
        get_daily_date_ranges(start_date=date, end_date=date + timedelta(days=1))
    )


def get_monthly_date_range_for_single_date(date: date) -> DateTimeRange:
    """
    Return single monthly date range for a given date.
    """
    first_day_of_month = date.replace(day=1)
    first_day_next_month = (date.replace(day=1) + timedelta(days=32)).replace(day=1)
    return date_range_to_date_time_range(
        DateRange(from_date=first_day_of_month, to_date=first_day_next_month)
    )


def get_weekly_date_range_for_single_date(date: date) -> DateTimeRange:
    """
    Return single weekly date range for a given date.
    """
    first_day_of_week = date - timedelta(days=date.weekday())
    first_day_next_week = first_day_of_week + timedelta(days=7)
    return date_range_to_date_time_range(
        DateRange(from_date=first_day_of_week, to_date=first_day_next_week)
    )



"""
Module to test telemetry mixin class for pipeline telemetry module.
"""

from test_data import DEFAULT_TELEMETRY_PARAMS

import pipeline_telemetry.mixin as MX
from pipeline_telemetry.main import Telemetry


def test_telemetry_mixin_exists():
    """check that TelemetryMixin class exists"""
    assert MX.TelemetryMixin


def test_set_source_name():
    """
    test method set_telemetry_source_name changes the telemetry
    source name attribute
    """

    class TelTest(MX.TelemetryMixin):
        def __init__(self):
            self._telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS)

        def get_source_name(self):
            return self._telemetry.telemetry.source_name

    tel_test = TelTest()
    test_source_name = "abcd"
    assert not tel_test.get_source_name() == test_source_name
    tel_test.set_telemetry_source_name(test_source_name)
    assert tel_test.get_source_name() == test_source_name


def test_process_errors_from_return_value(mocker):
    """
    Test method process_errors_from_return_value calls helper method
    add_errors_from_return_value
    """
    mocker.patch(
        ("pipeline_telemetry.mixin." "add_errors_from_return_value"), return_value=None
    )
    add_errors_spy = mocker.spy(MX, "add_errors_from_return_value")

    MX.TelemetryMixin().process_errors_from_return_value(
        sub_process="TEST", return_value="test"
    )
    assert add_errors_spy.called


def test_process_telemetry_counters_from_return_value(mocker):
    """
    Test method process_telemetry_counters_from_return_value calls helper
    method add_telemetry_counters_from_return_value
    """
    mocker.patch(
        ("pipeline_telemetry.mixin." "add_telemetry_counters_from_return_value"),
        return_value="correct return object",
    )
    add_telemetry_counters_spy = mocker.spy(
        MX, "add_telemetry_counters_from_return_value"
    )

    result = MX.TelemetryMixin().process_telemetry_counters_from_return_value(
        return_value="test"
    )
    assert add_telemetry_counters_spy.called
    assert result == "correct return object"


def test_process_telemetry_counters_from_result_list(mocker):
    """
    Test method process_telemetry_counters_from_result_list calls mixin
    method process_telemetry_counters_from_return_value
    """
    mocker.patch(
        (
            "pipeline_telemetry.mixin.TelemetryMixin."
            "process_telemetry_counters_from_return_value"
        ),
        return_value="correct return object",
    )
    process_telemetry_counters_spy = mocker.spy(
        MX.TelemetryMixin, "process_telemetry_counters_from_return_value"
    )

    result = MX.TelemetryMixin().process_telemetry_counters_from_list(
        result_list=["test"]
    )
    assert process_telemetry_counters_spy.called
    assert result == "correct return object"



from datetime import date, datetime, timedelta
from typing import Iterator, List

from test_data import DEFAULT_TELEMETRY_MODEL_PARAMS, DEFAULT_TELEMETRY_PARAMS

from pipeline_telemetry import TelemetrySelector
from pipeline_telemetry.data_classes import TelemetryData, TelemetryModel

TEST_TELEMETRY_SELECTOR = TelemetrySelector(**DEFAULT_TELEMETRY_PARAMS)

TODAY = date.today()
YESTERDAY = date.today() - timedelta(days=1)
DAY_BEFORE_YESTERDAY = date.today() - timedelta(days=2)

TELEMETRY_MODEL_1 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)

telemetry_model_2 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
telemetry_data_2 = TelemetryData(base_counter=1)
telemetry_model_2.telemetry.update({"DATA_STORAGE": telemetry_data_2})

telemetry_model_3 = TelemetryModel(**DEFAULT_TELEMETRY_MODEL_PARAMS)
telemetry_data_3 = TelemetryData(base_counter=2, fail_counter=1)
telemetry_model_3.telemetry.update({"DATA_STORAGE": telemetry_data_3})


class TelemetryTestList:
    __telemetry_models: List[TelemetryModel] = [
        TELEMETRY_MODEL_1,
        telemetry_model_2,
        telemetry_model_3,
    ]
    stored_telemetry: List[TelemetryModel]

    def telemetry_list(
        self,
        telemetry_type: str,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: str,
        from_date_time: datetime,
        to_date_time: datetime,
    ) -> Iterator[TelemetryModel]:
        """
        Method to return an iteraror TelemetryModel instances retrieved from a database
        query with the provided arguments.
        """
        for telemetry_model in self.__telemetry_models:
            yield telemetry_model

    def store_telemetry(self, telemetry: TelemetryModel) -> None:
        """public method to persist telemetry object"""
        if not hasattr(self, "stored_telemetry"):
            self.stored_telemetry = []
        self.stored_telemetry.append(telemetry)



"""Module to define errors"""

from errors import ErrorCode
from errors.base import FunctionalErrorsBaseClass


class ValidationErrors(FunctionalErrorsBaseClass):
    """
    Class to define enumerator errors for Error module
    """

    KEY_NOT_FOUND = ErrorCode(
        code="HAS_KEY_ERR_0001", description="Key missing in provided dict"
    )

    FIELD_NOT_FOUND = ErrorCode(
        code="VALIDATE_ENTRIES_ERR_001",
        description="Field that needs counting can not be found",
    )

    WRONG_TYPE_IN_FIELD = ErrorCode(
        code="VALIDATE_ENTRIES_ERR_002",
        description="Field does not contain dict or list",
    )

    UNEXPECTED_NR_OF_ITEMS = ErrorCode(
        code="VALIDATE_ENTRIES_ERR_003",
        description="Field contains unexpected nr of items",
    )

    ENTRIES_FIELD_NOT_FOUND = ErrorCode(
        code="ENTRIES_HAVE_KEY_ERR_001", description="Entries field does not exist"
    )

    ENTRIES_FIELD_OF_WRONG_TYPE = ErrorCode(
        code="ENTRIES_HAVE_KEY_ERR_002",
        description="Entries field does not contain a list",
    )

    KEY_NOT_FOUND_IN_ENTRY = ErrorCode(
        code="ENTRIES_HAVE_KEY_ERR_003", description="Key missing in entry"
    )

    ENTRY_IS_NOT_A_DICT = ErrorCode(
        code="ENTRIES_HAVE_KEY_ERR_004",
        description="Entry that needs to have a key is not a dict",
    )



"""Module to test the momgo Aggregators."""

from test_aggregator_data import TEST_TELEMETRY_SELECTOR

from pipeline_telemetry import DailyAggregator, DailyMongoAggregator
from pipeline_telemetry.storage import TelemetryMongoStorage


def test_daily_mongo_aggregator_class_exists():
    """Test that DailyMongoAggregator class exists."""
    assert DailyMongoAggregator


def test_daily_mongo_aggregator_returns_daily_telemerty_class():
    """
    Test that DailyMongoAggregator class when instanciated returns
    a DailyAggregator with TelemetryMongoStorage class.
    """
    aggregator = DailyMongoAggregator(TEST_TELEMETRY_SELECTOR)
    assert isinstance(aggregator, DailyAggregator)
    assert isinstance(aggregator.storage_class, TelemetryMongoStorage)



"""
module to test validate module of data-validator
"""

import pytest
from test_data import InstructionTestClass

from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.validators.dict_validator import DictValidator
from pipeline_telemetry.validators.has_key import HasKey

# pylint: disable=protected-access


def test_dict_validator_class_exists():
    """check that DictValidator class exists"""
    assert DictValidator


def test_dict_validator_has_instructions():
    """check that DictValidator class has instructions attribute"""
    assert DictValidator._instructions


def test_dict_validator_register_instruction():
    """
    check that register_instruction method registers an instruction
    """
    # ensure that test is not corrupted by previous registrations
    DictValidator._instructions = {}
    DictValidator.register_instruction(InstructionTestClass)
    assert InstructionTestClass.INSTRUCTION in list(DictValidator._instructions.keys())


def test_dict_validator_register_instruction_twice():
    """
    check that register_instruction method raises exception if you try to
    register an instruction twice
    """
    # ensure that test is not corrupted by previous registrations
    DictValidator._instructions = {}
    DictValidator.register_instruction(InstructionTestClass)
    with pytest.raises(exceptions.InstructionRegisteredTwice):
        DictValidator.register_instruction(InstructionTestClass)


def test_apply_rule_raises_exception_with_unregistered_instruction():
    """
    test that _apply_rule method returns an exception when a rule with an
    unregistered instruction is provided
    """
    with pytest.raises(exceptions.UnknownInstruction):
        DictValidator()._apply_rule({}, ("unkwown_rule", {"rule details": "details"}))


def test_apply_rule_calls_do_validate_method(mocker):
    """
    test that _apply_rule method calls the `validate` method on the instruction
    class and returns the value receive from the validate method
    """
    # as previous tests might have cleaned the instruction set it needs to be
    # reregistered
    DictValidator.register_instruction(HasKey)
    rule = ("has_key", {"has_key_details": "rule"})
    mocker.patch(
        "pipeline_telemetry.validators.has_key.HasKey.validate",
        return_value="return_value",
    )
    _do_validate_spy = mocker.spy(HasKey, "validate")
    return_value = DictValidator()._apply_rule({}, rule)
    assert return_value == "return_value"
    assert _do_validate_spy.called


def test_instruction_from_rule_succes():
    """test instruction returns instruction when given a valid rule"""
    valid_rule = ("instruction_in_valid_rule", {"rule_details": "test"})
    assert (
        DictValidator._instruction_from_rule(valid_rule) == "instruction_in_valid_rule"
    )


def test_instruction_from_rule_with_more_the_one_rule():
    """test instruction returns instruction when given a valid rule"""
    to_many_rules = {"rule_1": "test1", "rule_2": "test2"}
    with pytest.raises(exceptions.RuleCanHaveOnlyOneInstruction):
        DictValidator._instruction_from_rule(to_many_rules)



"""Module to define validate entries validator class"""

from dataclasses import dataclass
from typing import Optional

import jmespath
from errors import ErrorCode, ListErrors, add_error_data

from ..settings import exceptions
from .abstract_validator_instruction import (
    AbstractValidatorInstruction,
    BaseValidatorInstructionRuleData,
)


@dataclass(frozen=True)
class ValidateEntriesRuleData(BaseValidatorInstructionRuleData):
    expected_count: int

    def __post_init__(self):
        if not isinstance(self.expected_count, int):
            raise TypeError("Field 'expected_count' must be of type 'int'.")

        if self.expected_count < 0:
            raise exceptions.ExpectedCountMustBePositiveInt

        super().__post_init__()


class ValidateEntries(AbstractValidatorInstruction):
    """
    class to define validate entries instruction.
    This instruction counts the entries in a list or a dict and reports an
    error when the number of items in the list or dict is different from the
    expected nr of items
    """

    RULE_DATA_CLASS = ValidateEntriesRuleData
    INSTRUCTION = "validate_entries"
    expected_count_field = "expected_count"

    @classmethod
    def _validate(
        cls, dict_to_validate: dict, rule_data: BaseValidatorInstructionRuleData
    ) -> list[ErrorCode]:
        """
        method to do the actual validation

        returns:
            - error in case
        """
        fieldname = cls._get_field_name(rule_data)

        field_to_validate = jmespath.search(fieldname, dict_to_validate)

        if not field_to_validate:
            return cls._validation_error(ListErrors.FIELD_NOT_FOUND, fieldname)

        if not isinstance(field_to_validate, (list, dict)):
            return cls._validation_error(ListErrors.WRONG_TYPE_IN_FIELD, fieldname)

        if len(field_to_validate) != cls._get_expected_count(rule_data):
            return cls._validation_error(ListErrors.UNEXPECTED_NR_OF_ITEMS, fieldname)

        return []

    @classmethod
    def _get_expected_count(
        cls, rule_data: BaseValidatorInstructionRuleData
    ) -> Optional[str]:
        """
        retrieves the expected_count in scope of the rule content
        """
        return getattr(rule_data, cls.expected_count_field)

    @staticmethod
    def _validation_error(error_code: ErrorCode, fieldname: str) -> list[ErrorCode]:
        """returns error code object in list with fieldname as error data"""
        return [add_error_data(error=error_code, error_data=fieldname)]



"""Module to provide a storage class for using MongoDB.

When using MongoDB storage class you can select the mongoDB instance to be used
via environment Variables

MONGO_DB_NAME (defaults to `telemetry`)
MONGO_DB_PASSWORD
MONGO_DB_USERNAME
MONGO_DB_HOST
MONGO_DB_PORT

If no host and port are defined the connection will dedault to a localhost
mongoDB instance.
"""

from datetime import datetime
from typing import Any, Dict, Iterator

from mongoengine import (
    DateTimeField,
    DictField,
    Document,
    FloatField,
    StringField,
    connect,
)

from ..data_classes import TelemetryModel
from ..settings import settings as st
from .generic import AbstractTelemetryStorage
from .mongo_connection import MONGO_ACCESS_PARAMS

connect(alias="telemetry", **MONGO_ACCESS_PARAMS)


class TelemetryMongoModel(Document):
    """
    Class to provice telemetry Mongo Model for persistance in MongoDB
    """

    category = StringField()
    sub_category = StringField()
    source_name = StringField()
    process_type = StringField()
    start_date_time = DateTimeField()
    run_time_in_seconds = StringField()
    traffic_light = StringField()
    telemetry_type = StringField()
    io_time_in_seconds = FloatField(default=0)
    telemetry = DictField(default=None)

    meta = {
        "db_alias": "telemetry",
        "indexes": [
            "category",
            "sub_category",
            "source_name",
            ("category", "sub_category", "source_name", "process_type"),
            "process_type",
            "traffic_light",
            "start_date_time",
        ],
    }

    def to_dict(self) -> Dict:
        """
        Method to convert TelemetryMongoModel instance to a dict that can be
        used to instatiate a TelemetryModel object.
        The mongo object '_id' field needs to be removed as it is not used by
        TelemetryModel.
        """
        telemetry_dict = self.to_mongo().to_dict()
        telemetry_dict.pop("_id", None)
        # run_time_on_seconds is stored as str in Mongo but needs to be a float
        # when processing telemetry data
        telemetry_dict[st.RUN_TIME] = float(telemetry_dict[st.RUN_TIME])
        return telemetry_dict


class TelemetryMongoStorage(AbstractTelemetryStorage):
    """
    Class to provice telemetry in mongo storage class.
    This class can be used as storage_class argument when creating
    an instance of Telemetry.
    """

    def store_aggregated_telemetry(self, telemetry: TelemetryModel) -> None:
        """public method to persist telemetry object"""
        self._remove_existing_aggregation_telemetry(telemetry)
        self.store_telemetry(telemetry)

    def store_telemetry(self, telemetry: TelemetryModel) -> None:
        """public method to persist telemetry object"""
        telemetry_mongo_kwargs = self._telemetry_model_kwargs(telemetry)
        TelemetryMongoModel(**telemetry_mongo_kwargs).save()

    def _remove_existing_aggregation_telemetry(self, telemetry: TelemetryModel) -> None:
        """
        Removes any already existing aggregations for a specific telemetry
        aggregation.
        If you want to run and store a new aggregation object (for example a
        daily aggrgation) then the already daily aggregation for that day must
        be removed.

        Args:
            telemetry (TelemetryModel): The new telemetry aggregation object
        """
        query_params_exist_aggr = self._get_aggr_telem_query_params(telemetry)
        for telemetry_inst in self.select_records(**query_params_exist_aggr):
            telemetry_inst.delete()

    @staticmethod
    def _telemetry_model_kwargs(telemetry: TelemetryModel) -> dict:
        """
        Returns a dicts with kwargs that can be used to create a new
        TelemetryMongoStorage instance.
        """
        telemetry_type = getattr(telemetry, st.TELEMETRY_TYPE_KEY)
        category = getattr(telemetry, st.CATEGORY_KEY)
        sub_category = getattr(telemetry, st.SUB_CATEGORY_KEY)
        source_name = getattr(telemetry, st.SOURCE_NAME_KEY)
        process_type = getattr(telemetry, st.PROCESS_TYPE_KEY)
        start_date_time = getattr(telemetry, st.START_TIME)
        run_time_in_seconds = getattr(telemetry, st.RUN_TIME)
        traffic_light = getattr(telemetry, st.TRAFFIC_LIGHT_KEY)
        io_time_in_seconds = getattr(telemetry, st.IO_TIME_KEY)
        telemetry_data = {
            k: v.__dict__ for k, v in getattr(telemetry, st.TELEMETRY_FIELD_KEY).items()
        }

        return {
            st.TELEMETRY_TYPE_KEY: telemetry_type,
            st.CATEGORY_KEY: category,
            st.SUB_CATEGORY_KEY: sub_category,
            st.SOURCE_NAME_KEY: source_name,
            st.PROCESS_TYPE_KEY: process_type,
            st.START_TIME: start_date_time,
            st.RUN_TIME: str(round(run_time_in_seconds, 2)),
            st.TRAFFIC_LIGHT_KEY: traffic_light,
            st.TELEMETRY_FIELD_KEY: telemetry_data,
            st.IO_TIME_KEY: io_time_in_seconds,
        }

    def select_records(
        self,
        telemetry_type: str,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: str,
        from_date_time: datetime,
        to_date_time: datetime,
    ) -> Iterator:
        """
        Select telemetry records unique to a single process, source category
        and sub category for as specific time period.
        """
        query_details = {
            "telemetry_type": telemetry_type,
            "category": category,
            "sub_category": sub_category,
            "source_name": source_name,
            "process_type": process_type,
        }

        return TelemetryMongoModel.objects(
            start_date_time__gte=from_date_time,
            start_date_time__lt=to_date_time,
            **query_details,
        )

    @staticmethod
    def _db_object_to_dict(db_object: Any) -> Dict:
        """Returns a db object as a dict object."""
        # Mongo onbject need to be converted to Dict object first
        return db_object.to_dict()



"""Module to run integration test for telemetry module"""

from test_data import DEFAULT_TELEMETRY_PARAMS
from test_integration_data import (
    BASIC_RETRIEVE_DATA_TEST,
    ITEMS_FIELD_HAS_DICT_TEST,
    MISSING_KEY_RETRIEVE_DATA_TEST,
    MUST_HAVE_KEY_IN_ITEMS_TEST,
    NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST,
)

from pipeline_telemetry.main import Telemetry


def test_basic_test_data():
    """Test data basic data test returns correct telemetry."""
    telemetry_rules = BASIC_RETRIEVE_DATA_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | {"telemetry_rules": telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry.add("RETRIEVE_RAW_DATA", BASIC_RETRIEVE_DATA_TEST.data, None)
    telemetry.save_and_close()
    assert BASIC_RETRIEVE_DATA_TEST.validate_telemetry_data(telemetry)


def test_missing_key_integration_test():
    """Test telemetry returns correct telemetry object when key is missing."""
    telemetry_rules = MISSING_KEY_RETRIEVE_DATA_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | {"telemetry_rules": telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry.add("RETRIEVE_RAW_DATA", MISSING_KEY_RETRIEVE_DATA_TEST.data, None)
    telemetry.save_and_close()
    assert MISSING_KEY_RETRIEVE_DATA_TEST.validate_telemetry_data(telemetry)


def test_must_have_key_in_items_integration_test():
    """
    Test telemetry returns correct telemetry object when key is missing in one
    of the items.
    """
    telemetry_rules = MUST_HAVE_KEY_IN_ITEMS_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | {"telemetry_rules": telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry.add("RETRIEVE_RAW_DATA", MUST_HAVE_KEY_IN_ITEMS_TEST.data, None)
    telemetry.save_and_close()
    assert MUST_HAVE_KEY_IN_ITEMS_TEST.validate_telemetry_data(telemetry)


def test_no_items_have_must_have_key_in_items_integration_test():
    """
    Test telemetry returns correct telemetry object when no item has the
    correct key.
    """
    telemetry_rules = NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | {"telemetry_rules": telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry.add("RETRIEVE_RAW_DATA", NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST.data, None)
    telemetry.save_and_close()
    assert NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST.validate_telemetry_data(telemetry)


def test_items_has_dict_field_test():
    """
    Test telemetry returns correct telemetry object when items field contains
    a dict field.
    """
    telemetry_rules = ITEMS_FIELD_HAS_DICT_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | {"telemetry_rules": telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry.add("RETRIEVE_RAW_DATA", ITEMS_FIELD_HAS_DICT_TEST.data, None)
    telemetry.save_and_close()
    assert ITEMS_FIELD_HAS_DICT_TEST.validate_telemetry_data(telemetry)



"""Module to test the DailyAggregator class."""

from test_aggregator_data import TelemetryTestList

from pipeline_telemetry import PartialToSingleAggregator, TelemetrySelector
from pipeline_telemetry.data_classes import TelemetryModel
from pipeline_telemetry.settings.date_ranges import get_daily_date_range_yesterday

keys_in_telemetry_list_params = [
    "telemetry_type",
    "category",
    "sub_category",
    "source_name",
    "process_type",
    "from_date_time",
    "to_date_time",
]


def test_partial_to_daily_aggregator_exists():
    """Test that the PartialToDailyAggregator class exists."""
    assert PartialToSingleAggregator


def test_run_aggregation():
    """Method to test a single aggregation."""
    telemetry_selector = TelemetrySelector(
        category="test",
        sub_category="sub_test",
        source_name="source",
        process_type="process_type",
    )
    aggr = PartialToSingleAggregator(
        telemetry_selector=telemetry_selector, telemetry_storage=TelemetryTestList()
    )
    result_telemetry = aggr._run_aggregation(next(get_daily_date_range_yesterday()))
    assert isinstance(result_telemetry, TelemetryModel)
    assert result_telemetry.telemetry["DATA_STORAGE"].base_counter == 3
    assert result_telemetry.telemetry["DATA_STORAGE"].fail_counter == 1



"""Module to test aggregator logic."""

from pipeline_telemetry import TelemetryAggregator


def test_aggegator_class_exists():
    """Test the TelemetryAggregator class exists."""
    assert TelemetryAggregator


def test_initializing_aggegator_instance():
    """
    Test that the TelemetryAggregator can be instantiated with an int as int support the __add__ method
    """
    assert TelemetryAggregator(1)


def test_aggegator_from_adds_all_telemetry_list_items():
    """
    Test that the aggregate_from method adds all items from a telemetry list to
    the telemetry object and returns the result.
    """
    new_telemetry = 1
    aggregator = TelemetryAggregator(new_telemetry)
    telemetry_list = [2, 3, 4]
    assert aggregator.aggregate(telemetry_list) == 10



"""Module for testing date_ranges functions."""

from datetime import date, datetime, timedelta

from pipeline_telemetry.settings.date_ranges import (
    DateTimeRange,
    date_range_generator,
    get_daily_date_range_for_single_date,
    get_daily_date_range_till_yesterday,
    get_daily_date_range_yesterday,
    get_daily_date_ranges,
    get_monthly_date_range_for_single_date,
    get_weekly_date_range_for_single_date,
)


def str_to_date_time(date_str: str) -> datetime:
    return datetime.strptime(date_str, "%Y-%m-%d")


def str_to_date(date_str: str) -> date:
    return str_to_date_time(date_str).date()


def test_date_range_generator():
    """Test date_range_generator generates DateTimeRanges."""
    start_date_time = datetime.strptime("2022-01-18", "%Y-%m-%d")
    time_delta = timedelta(days=1)
    end_date_time = start_date_time + timedelta(days=3)
    daterange_generator = date_range_generator(
        start_date_time=start_date_time,
        end_date_time=end_date_time,
        time_delta=time_delta,
    )
    result = [daterange for daterange in daterange_generator]
    assert result == [
        DateTimeRange(str_to_date_time("2022-01-18"), str_to_date_time("2022-01-19")),
        DateTimeRange(str_to_date_time("2022-01-19"), str_to_date_time("2022-01-20")),
        DateTimeRange(str_to_date_time("2022-01-20"), str_to_date_time("2022-01-21")),
    ]


def test_date_range_generator_with_during_day_time_stamps():
    """
    Test date_range_generator generates correct DateTimeRanges when atributes
    contain a time other the 00:00:00. The first daterange should start with
    the date part of the from_data."""
    start_date_time = datetime.strptime("2022-01-18 13:55:26", "%Y-%m-%d %H:%M:%S")
    time_delta = timedelta(days=1)
    end_date_time = start_date_time + timedelta(days=3)
    daterange_generator = date_range_generator(
        start_date_time=start_date_time,
        end_date_time=end_date_time,
        time_delta=time_delta,
    )
    result = [daterange for daterange in daterange_generator]
    assert result == [
        DateTimeRange(str_to_date_time("2022-01-18"), str_to_date_time("2022-01-19")),
        DateTimeRange(str_to_date_time("2022-01-19"), str_to_date_time("2022-01-20")),
        DateTimeRange(str_to_date_time("2022-01-20"), str_to_date_time("2022-01-21")),
    ]


def test_get_daily_date_ranges():
    """Test daterange_generator returns DateTimeRange ranges."""
    start_date = date(year=2022, month=1, day=18)
    end_date = date(year=2022, month=1, day=21)
    daterange_generator = get_daily_date_ranges(
        start_date=start_date, end_date=end_date
    )
    result = [daterange for daterange in daterange_generator]
    assert result == [
        DateTimeRange(str_to_date_time("2022-01-18"), str_to_date_time("2022-01-19")),
        DateTimeRange(str_to_date_time("2022-01-19"), str_to_date_time("2022-01-20")),
        DateTimeRange(str_to_date_time("2022-01-20"), str_to_date_time("2022-01-21")),
    ]


def test_get_daily_date_range_till_yesterday():
    """
    Test get_daily_date_range_till_yesterday generates a DateTimeRange
    till yesterday.
    """
    start_date = date.today() - timedelta(days=3)
    daterange_generator = get_daily_date_range_till_yesterday(start_date=start_date)
    result = [daterange for daterange in daterange_generator]
    assert len(result) == 3


def test_get_daily_date_ranges_yesterday():
    """
    Test get_daily_date_ranges_yesterday generates a DateTimeRange
    for only yesterday.
    """
    today = datetime.strftime(datetime.now(), "%Y-%m-%d")
    yesterday = datetime.strftime(datetime.now() - timedelta(days=1), "%Y-%m-%d")

    daterange_generator = get_daily_date_range_yesterday()
    result = [daterange for daterange in daterange_generator]
    assert result[0].from_date == str_to_date_time(yesterday)
    assert result[0].to_date == str_to_date_time(today)
    assert len(result) == 1


def test_get_daily_date_range_for_single_date():
    """
    Test get_daily_date_range_for_single_date returns a single DateTimeRange.
    """
    today = date.today()
    yesterday = date.today() - timedelta(days=1)
    single_daily_date_range = get_daily_date_range_for_single_date(date=yesterday)

    assert isinstance(single_daily_date_range, DateTimeRange)
    assert single_daily_date_range.from_date.date() == yesterday
    assert single_daily_date_range.to_date.date() == today


def test_get_monthly_date_range_for_single_date():
    """
    Test get_monthly_date_range_for_single_date returns a single DateTimeRange.
    """
    first_day_of_month = datetime.strptime("2022-02-01", "%Y-%m-%d")
    first_day_of_next_month = datetime.strptime("2022-03-01", "%Y-%m-%d")
    input_date = datetime.strptime("2022-02-11", "%Y-%m-%d")
    single_monthly_date_range = get_monthly_date_range_for_single_date(date=input_date)
    assert isinstance(single_monthly_date_range, DateTimeRange)
    assert single_monthly_date_range.from_date == first_day_of_month
    assert single_monthly_date_range.to_date == first_day_of_next_month


def test_get_weekly_date_range_for_single_date():
    """
    Test get_monthly_date_range_for_single_date returns a single DateTimeRange.
    """
    first_day_of_week = datetime.strptime("2022-10-03", "%Y-%m-%d")
    first_day_of_next_week = datetime.strptime("2022-10-10", "%Y-%m-%d")
    input_date = datetime.strptime("2022-10-07", "%Y-%m-%d")
    single_weekly_date_range = get_weekly_date_range_for_single_date(
        date=input_date.date()
    )
    assert isinstance(single_weekly_date_range, DateTimeRange)
    assert single_weekly_date_range.from_date == first_day_of_week
    assert single_weekly_date_range.to_date == first_day_of_next_week



from .helper import TelemetryAggregator, TelemetrySelector
from .mongo_aggregator import (
    DailyAggregator,
    DailyMongoAggregator,
    PartialToSingleAggregator,
    PartialToSingleMongoAggregator,
)

__all__ = [
    "TelemetryAggregator",
    "TelemetrySelector",
    "DailyAggregator",
    "DailyMongoAggregator",
    "PartialToSingleAggregator",
    "PartialToSingleMongoAggregator",
]



"""
Module to provide Aggregator classes with predefined TelemetryMongoStorage
as storage class.

All these classes only require a TelemetrySelector upon instantiation and
return the appropriate Aggregator class with TelemetryMongoStorage as storage
class.

>>> telemetry_selector = TelemetrySelector(
        category, sub_category, source_name, process_type)
>>> aggeregator = MongoAggregatorClassName(
            telemetry_selector=telemetry_selector
        )

available classes:
- DailyMongoAggregator

"""

from abc import ABC

from pipeline_telemetry.storage import TelemetryMongoStorage

from .aggregator import AbstractAggregator, DailyAggregator, PartialToSingleAggregator
from .helper import TelemetrySelector


class AbstractMongoAggregator(ABC):
    """Class to return a DailyAggregator class with TelemetryMongoStorage.

    When using this aggregator class the storage_class is automatically set
    to TelemetryMongoStorage and can not be provided as argument.

    >>> telemetry_selector = TelemetrySelector(
        category, sub_category, source_name, process_type)
    >>> aggeregator = DailyMongoAggregator(
            telemetry_selector=telemetry_selector
        )
    """

    AGGREGATOR_CLASS: type[AbstractAggregator]

    def __new__(cls, telemetry_selector: TelemetrySelector):
        return cls.AGGREGATOR_CLASS(
            telemetry_selector=telemetry_selector,
            telemetry_storage=TelemetryMongoStorage(),
        )


class DailyMongoAggregator(AbstractMongoAggregator):
    """Class to return a DailyAggregator class with TelemetryMongoStorage.

    When using this aggregator class the storage_class is automatically set
    to TelemetryMongoStorage and can not be provided as argument.

    >>> telemetry_selector = TelemetrySelector(
        category, sub_category, source_name, process_type)
    >>> aggeregator = DailyMongoAggregator(
            telemetry_selector=telemetry_selector
        )
    """

    AGGREGATOR_CLASS = DailyAggregator


class PartialToSingleMongoAggregator(AbstractMongoAggregator):
    """
    Class to return a PartialToSingleAggregator class with
    TelemetryMongoStorage.

    When using this aggregator class the storage_class is automatically set
    to TelemetryMongoStorage and can not be provided as argument.
    """

    AGGREGATOR_CLASS = PartialToSingleAggregator



"""
module to test has_key instruction class
"""

from test_data import TEST_ERROR_CODE

from pipeline_telemetry.settings.telemetry_errors import ErrorCode
from pipeline_telemetry.validators.abstract_validator_instruction import (
    BaseValidatorInstructionRuleData,
)
from pipeline_telemetry.validators.has_key import HasKey

# pylint: disable=protected-access


def test_has_key_class_exists():
    """check that HasKey class exists"""
    assert HasKey


def test_has_key_class_is_singelton():
    """check that HasKey is a singleton class"""
    instance_one = HasKey()
    instance_two = HasKey()
    assert instance_one is instance_two


def test_get_field_name_method():
    """
    check that _get_field_name method returns value of
    field_name that is in scope of validation
    """
    rule_data = BaseValidatorInstructionRuleData(**{"field_name": "test"})
    assert HasKey._get_field_name(rule_data) == "test"


def test_validation_error_method():
    """
    Check that _validation_error method returns a list with
    ErrorCode instance.
    """
    validation_error_list = HasKey._validation_error(TEST_ERROR_CODE, "test_error_data")
    validation_error = validation_error_list[0]
    assert isinstance(validation_error_list, list)
    assert isinstance(validation_error, ErrorCode)


def test_validate_valid_dict():
    """Test that a dict with right key returns empty error list"""
    valid_dict = {"correct_key": "value"}
    rule_dict = {"field_name": "correct_key"}
    validation_result = HasKey.validate(
        dict_to_validate=valid_dict, rule_dict=rule_dict
    )
    assert validation_result == []


def test_validate_valid_dict_with_nested_key():
    """
    Test that a dict with right key returns empty error list when the nested
    key in input dict is found
    """
    valid_dict = {"correct_key": {"nested_key": "value"}}
    rule_dict = {"field_name": "correct_key.nested_key"}
    validation_result = HasKey.validate(
        dict_to_validate=valid_dict, rule_dict=rule_dict
    )
    assert validation_result == []


def test_validate_invalid_dict():
    """
    Test that a dict without right key returns a list with error_code
    KEY_NOT_FOUND_001 and the field_name in the error_data

    """
    invalid_dict = {"incorrect_key": "value"}
    rule_dict = {"field_name": "correct_key"}
    validation_result = HasKey.validate(
        dict_to_validate=invalid_dict, rule_dict=rule_dict
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == "HAS_KEY_ERR_0001@KEY_<correct_key>"


def test_validate_invalid_nested_dict():
    """
    Test that a nested dict without right nested key returns a list with
    error_code KEY_NOT_FOUND_001 and the nested field_name in the error_data.
    """
    invalid_dict = {"incorrect_key": {"nested_key": "value"}}
    rule_dict = {"field_name": "correct_key.nested_key"}
    validation_result = HasKey.validate(
        dict_to_validate=invalid_dict, rule_dict=rule_dict
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == "HAS_KEY_ERR_0001@KEY_<correct_key.nested_key>"



"""Module to test storage module."""

from datetime import datetime, timedelta

from test_data import DEFAULT_TELEMETRY_MODEL_PARAMS

from pipeline_telemetry.data_classes import TelemetryModel
from pipeline_telemetry.settings.settings import DEFAULT_TRAFIC_LIGHT_COLOR, RUN_TIME
from pipeline_telemetry.storage.mongo import TelemetryMongoModel, TelemetryMongoStorage
from pipeline_telemetry.storage.mongo_connection import get_mongo_db_port


def telemetry_query_params():
    """Returns a default set of query params"""
    return DEFAULT_TELEMETRY_MODEL_PARAMS | {
        "from_date_time": datetime.now() - timedelta(days=1),
        "to_date_time": datetime.now() + timedelta(days=1),
    }


def test_telemetry_mongo_model_class_exists():
    """Test TelemetryMongoModel class exists."""
    assert TelemetryMongoModel


def test_telemetry_mongo_storage_class_exists():
    """Test TelemetryMongoStorage class exists."""
    assert TelemetryMongoStorage


def test_store_telemetry_method_processes_and_saves_telemetry(mocker):
    """
    Test store_telemetry method calls _telemetry_model_kwargs and then creates
    and saves a mongo object.
    """
    mongo_module_path = "pipeline_telemetry.storage.mongo."
    mocker.patch(mongo_module_path + "TelemetryMongoModel.save", return_value=True)
    mocker.patch(
        mongo_module_path + "TelemetryMongoStorage._telemetry_model_kwargs",
        return_value={},
    )
    _telemetry_model_kwargs_spy = mocker.spy(
        TelemetryMongoStorage, "_telemetry_model_kwargs"
    )
    _telemetry_mongo_model_new_spy = mocker.spy(TelemetryMongoModel, "__new__")
    _telemetry_mongo_model_save_spy = mocker.spy(TelemetryMongoModel, "save")
    TelemetryMongoStorage().store_telemetry(telemetry={"source_name": "test"})
    assert _telemetry_model_kwargs_spy.called
    assert _telemetry_mongo_model_new_spy.called
    assert _telemetry_mongo_model_save_spy.called


def test_telemetry_model_kwargs_method():
    """Test _telemetry_model_kwargs method returns correct kwargs."""
    result = TelemetryMongoStorage()._telemetry_model_kwargs(
        TelemetryModel(
            **{
                "telemetry_type": "test telemetry type",
                "category": "test",
                "sub_category": "sub_test",
                "source_name": "tst_source_name",
                "process_type": "tst_process_type",
                "start_date_time": "tst_start_date_time",
                "run_time_in_seconds": 1.123,
                "io_time_in_seconds": 1.1,
                "traffic_light": DEFAULT_TRAFIC_LIGHT_COLOR,
            }
        )
    )

    assert result == {
        "telemetry_type": "test telemetry type",
        "category": "test",
        "sub_category": "sub_test",
        "source_name": "tst_source_name",
        "process_type": "tst_process_type",
        "start_date_time": "tst_start_date_time",
        "run_time_in_seconds": "1.12",
        "io_time_in_seconds": 1.1,
        "telemetry": {},
        "traffic_light": DEFAULT_TRAFIC_LIGHT_COLOR,
    }


def test_get_mongo_db_port_returns_none():
    """Test _get_mongo_db_port returns None if no env variable is set."""
    assert get_mongo_db_port() is None


def test_get_mongo_db_port_returns_port_int(mocker):
    """Test _get_mongo_db_port returns port nr as intno env variable is set."""
    mocker.patch("os.getenv", return_value="12345")
    assert get_mongo_db_port() == 12345


def test_default_access_params():
    """Test default_access_params are set poperly."""
    from pipeline_telemetry.storage import mongo_connection as mc

    def_params = mc.MONGO_ACCESS_PARAMS
    assert def_params["port"] == mc.DEFAULT_MONGO_DB_PORT
    assert def_params["host"] == mc.DEFAULT_MONGO_DB_HOST


def test_mongo_model_to_dict():
    """
    Test to_dict method returns a dict with run_time_in_seconds atrribute converted to float.
    """
    telemetry = TelemetryMongoModel()
    telemetry._id = "test _id"
    telemetry.run_time_in_seconds = "1"
    telemetry_to_dict = telemetry.to_dict()
    assert isinstance(telemetry_to_dict, dict)
    assert isinstance(getattr(telemetry, RUN_TIME), str)
    assert isinstance(telemetry_to_dict[RUN_TIME], float)
